@@ -91,7 +91,6 @@ class Conf:
         assert num_atoms == tot_atoms, "Conf() Error: missing hydrogens"
         self.natoms = num_atoms
         self.charge = rdmolops.GetFormalCharge(self.rdmol)
-        self.set_solvent_accessible_surface_area()
         self.props.update({
             'atoms': self.natoms, 
             'charge': self.charge,
@@ -195,6 +194,44 @@ class Conf:
         return np.sqrt(np.mean(b))
     
 
+    @property
+    def SASA(self) -> dict:
+        """Calculate Solvent Accessible Surface Area (total, polar, apolar).
+
+        Returns:
+            tuple[float, float, float]: (total, polar, apolar)
+        """
+        radii = [ptable.GetRvdw(atom.GetAtomicNum()) for atom in self.rdmol.GetAtoms()]
+        
+         # CalcSASA signature: CalcSASA(mol, radii, confId=conf_id) in modern RDKit
+        total_sasa = rdFreeSASA.CalcSASA(self.rdmol, radii=radii, confIdx=-1)
+
+        # try to read per-atom SASA values set by CalcSASA into atoms
+        atom_sasas = [atom.GetDoubleProp('SASA') for atom in self.rdmol.GetAtoms()]
+        
+        # Now compute polar/apolar sums: polar = N,O and Hs attached to them
+        polar_idx = set()
+        for atom in self.rdmol.GetAtoms():
+            if atom.GetAtomicNum() in (7, 8):  # N or O
+                polar_idx.add(atom.GetIdx())
+                # include attached hydrogens
+                for nbr in atom.GetNeighbors():
+                    if nbr.GetAtomicNum() == 1:
+                        polar_idx.add(nbr.GetIdx())
+
+        # default numeric 0 for None entries (defensive)
+        numeric_atom_sasas = [0.0 if v is None else float(v) for v in atom_sasas]
+        sasa_polar = sum(numeric_atom_sasas[i] for i in polar_idx)
+        sasa_total = sum(numeric_atom_sasas)
+        sasa_apolar = sasa_total - sasa_polar
+
+        return recursive_round({
+            'sasa_total': sasa_total, 
+            'sasa_polar': sasa_polar,
+            'sasa_apolar': sasa_apolar,
+        }, decimals=2)
+
+
     def copy(self) -> Self:
         """Returns a copy of self.
 
@@ -244,44 +281,6 @@ class Conf:
         
         return self
 
-    
-    def set_solvent_accessible_surface_area(self) -> None:
-        """Calculate Solvent Accessible Surface Area (total, polar, apolar).
-
-        Returns:
-            tuple[float, float, float]: (total, polar, apolar)
-        """
-        radii = [ptable.GetRvdw(atom.GetAtomicNum()) for atom in self.rdmol.GetAtoms()]
-        
-         # CalcSASA signature: CalcSASA(mol, radii, confId=conf_id) in modern RDKit
-        total_sasa = rdFreeSASA.CalcSASA(self.rdmol, radii=radii, confIdx=-1)
-
-        # try to read per-atom SASA values set by CalcSASA into atoms
-        atom_sasas = [atom.GetDoubleProp('SASA') for atom in self.rdmol.GetAtoms()]
-        
-        # Now compute polar/apolar sums: polar = N,O and Hs attached to them
-        polar_idx = set()
-        for atom in self.rdmol.GetAtoms():
-            if atom.GetAtomicNum() in (7, 8):  # N or O
-                polar_idx.add(atom.GetIdx())
-                # include attached hydrogens
-                for nbr in atom.GetNeighbors():
-                    if nbr.GetAtomicNum() == 1:
-                        polar_idx.add(nbr.GetIdx())
-
-        # default numeric 0 for None entries (defensive)
-        numeric_atom_sasas = [0.0 if v is None else float(v) for v in atom_sasas]
-        sasa_polar = sum(numeric_atom_sasas[i] for i in polar_idx)
-        sasa_total = sum(numeric_atom_sasas)
-        sasa_apolar = sasa_total - sasa_polar
-
-        self.props.update({
-            'sasa_total': sasa_total,
-            'sasa_polar': sasa_polar,
-            'sasa_apolar': sasa_apolar,
-        })
-
-
 
     def get_torsion_angle_atoms(self, strict: bool = True) -> list[tuple]:
         """Determine torsion/dihedral angle atoms (i-j-k-l) and rotating group for each rotatable bond (j-k).
