@@ -1369,6 +1369,15 @@ class Mol:
             return False
         
 
+    def get_SASA(self) -> dict:
+        """Get Solvent Accessible Surface Area"""
+        results = defaultdict(list)
+        for conf in self.confs:
+            for k, v in conf.SASA.items():
+                results[k].append(v)
+        return results
+
+
     def get_torsion_angle_atoms(self, strict: bool = True) -> list[tuple]:
         """Determine torsion/dihedral angle atoms (i-j-k-l) and rotating group for each rotatable bond (j-k).
 
