@@ -1,5 +1,6 @@
 import numpy as np
 
+from collections import deque
 from rdkit import Chem
 from rdworks.xtb.wrapper import GFN2xTB
 
@@ -34,11 +35,13 @@ def get_torsion_angle_atom_indices(rdmol:Chem.Mol, strict:bool=True) -> list[tup
         rotatable_bond_pattern = Chem.MolFromSmarts('[!$(*#*)&!D1]-&!@[!$(*#*)&!D1]')
     
     rotatable_bonds = rdmol.GetSubstructMatches(rotatable_bond_pattern)
-    
+    ri = rdmol.GetRingInfo()
+    atom_rings = ri.AtomRings()
+
     torsion_angle_atom_indices = []
 
     # small rings (n=3 or 4)
-    small_rings = [ r for r in list(rdmol.GetRingInfo().AtomRings()) if len(r) < 5 ]
+    small_rings = [ ring for ring in list(atom_rings) if len(ring) < 5 ]
     # ex. = [(1, 37, 35, 34, 3, 2), (29, 28, 30)]
 
     forbidden_terminal_nuclei = [1, 9, 17, 35, 53] # H,F,Cl,Br,I
@@ -94,6 +97,10 @@ def get_torsion_angle_atom_indices(rdmol:Chem.Mol, strict:bool=True) -> list[tup
         if is_in_small_ring:
             continue
 
+        # check ring closure between a and d - avoid macrocycles
+        if any(a_idx in ring and d_idx in ring for ring in atom_rings):
+            continue
+
         # determine a group of atoms to be rotated
         # https://ctr.fandom.com/wiki/Break_rotatable_bonds_and_report_the_fragments
         em = Chem.EditableMol(rdmol)
@@ -485,4 +492,57 @@ def create_torsion_fragment(rdmol: Chem.Mol,
 
     frag_indices = get_fragment_idx(rdmol, (i, j, k, l), frag_multi_breaks)
     
-    return (frag_multi_breaks, frag_indices, True, WBO_filtered)
\ No newline at end of file
+    return (frag_multi_breaks, frag_indices, True, WBO_filtered)
+
+
+def find_shortest_path_excluding_torsion(rdmol:Chem.Mol, i:int, j:int, k:int, l:int) -> list | None:
+    """
+    Find shortest path between atoms i and l, excluding the torsion path i-j-k-l.
+    
+    Uses BFS (Breadth-First Search) to find the shortest alternative path.
+    
+    Returns:
+        List of atom indices forming the path, or None if no path exists
+    """
+    # Get the bonds to exclude (the torsion path)
+    excluded_bonds = set()
+    bond_ij = rdmol.GetBondBetweenAtoms(i, j)
+    bond_jk = rdmol.GetBondBetweenAtoms(j, k)
+    bond_kl = rdmol.GetBondBetweenAtoms(k, l)
+    
+    if bond_ij:
+        excluded_bonds.add(bond_ij.GetIdx())
+    if bond_jk:
+        excluded_bonds.add(bond_jk.GetIdx())
+    if bond_kl:
+        excluded_bonds.add(bond_kl.GetIdx())
+    
+    # BFS to find shortest path from i to l
+    queue = deque([(i, [i])])  # (current_atom, path)
+    visited = {i}
+    
+    while queue:
+        current, path = queue.popleft()
+        
+        # Check neighbors
+        atom = rdmol.GetAtomWithIdx(current)
+        for neighbor in atom.GetNeighbors():
+            neighbor_idx = neighbor.GetIdx()
+            
+            # Get the bond between current and neighbor
+            bond = rdmol.GetBondBetweenAtoms(current, neighbor_idx)
+            
+            # Skip if this bond is part of the torsion path
+            if bond and bond.GetIdx() in excluded_bonds:
+                continue
+            
+            # Found the target
+            if neighbor_idx == l:
+                return path + [l]
+            
+            # Continue BFS
+            if neighbor_idx not in visited:
+                visited.add(neighbor_idx)
+                queue.append((neighbor_idx, path + [neighbor_idx]))
+    
+    return None  # No alternative path found
\ No newline at end of file
