@@ -6,6 +6,9 @@ import tempfile
 import logging
 import shutil
 import re
+import numpy as np
+from scipy.spatial import cKDTree
+from scipy.interpolate import griddata
 
 from pathlib import Path
 from types import SimpleNamespace
@@ -26,6 +29,165 @@ hartree2kcalpermol = 627.50947337481
 ev2kcalpermol = 23.060547830619026
 
 
+pt = Chem.GetPeriodicTable()
+
+
+def read_xyz_file(filename):
+    """Read molecular structure from XYZ file."""
+    with open(filename, 'r') as f:
+        lines = f.readlines()
+    
+    natoms = int(lines[0].strip())
+    atoms = []
+
+    for i in range(2, 2 + natoms):
+        parts = lines[i].split()
+        element = parts[0]
+        x, y, z = float(parts[1]), float(parts[2]), float(parts[3])
+        try:
+            atomic_num = pt.GetAtomicNumber(element)
+        except:
+            atomic_num = 6
+        atoms.append([atomic_num, x, y, z])
+    
+    return atoms
+
+
+def read_surface_esp(filename):
+    """
+    Read xtb ESP surface file with format: x y z esp_value
+    """
+    points = []
+    with open(filename, 'r') as f:
+        for line in f:
+            parts = line.split()
+            if len(parts) >= 4:
+                try:
+                    x = float(parts[0])
+                    y = float(parts[1])
+                    z = float(parts[2])
+                    esp = float(parts[3])
+                    points.append([x, y, z, esp])
+                except ValueError:
+                    continue
+    
+    return np.array(points)
+
+
+def create_regular_grid(surface_points, atoms, grid_spacing=0.6, padding=4.0):
+    """
+    Create a regular 3D grid encompassing the molecule.
+    
+    Args:
+        surface_points: Nx4 array of surface points (x, y, z, esp)
+        atoms: List of atomic coordinates
+        grid_spacing: Spacing between grid points (in Bohr or Angstrom)
+        padding: Extra space around molecule
+    """
+    # Get molecule bounds
+    all_coords = np.array([[a[1], a[2], a[3]] for a in atoms])
+    
+    # Determine grid bounds with padding
+    min_coords = np.min(all_coords, axis=0) - padding
+    max_coords = np.max(all_coords, axis=0) + padding
+    
+    # Also consider surface points
+    surface_coords = surface_points[:, :3]
+    min_coords = np.minimum(min_coords, np.min(surface_coords, axis=0) - padding)
+    max_coords = np.maximum(max_coords, np.max(surface_coords, axis=0) + padding)
+    
+    # Create grid
+    x_range = np.arange(min_coords[0], max_coords[0] + grid_spacing, grid_spacing)
+    y_range = np.arange(min_coords[1], max_coords[1] + grid_spacing, grid_spacing)
+    z_range = np.arange(min_coords[2], max_coords[2] + grid_spacing, grid_spacing)
+    
+    nx, ny, nz = len(x_range), len(y_range), len(z_range)
+    
+    logger.info(f"Grid dimensions: {nx} x {ny} x {nz} = {nx*ny*nz} points")
+    logger.info(f"Grid spacing: {grid_spacing:.3f}")
+    logger.info(
+        f"Grid bounds: X[{min_coords[0]:.2f}, {max_coords[0]:.2f}], "
+        f"Y[{min_coords[1]:.2f}, {max_coords[1]:.2f}], "
+        f"Z[{min_coords[2]:.2f}, {max_coords[2]:.2f}]")
+    
+    # Create 3D meshgrid
+    X, Y, Z = np.meshgrid(x_range, y_range, z_range, indexing='ij')
+    grid_points = np.c_[X.ravel(), Y.ravel(), Z.ravel()]
+    
+    origin = [x_range[0], y_range[0], z_range[0]]
+    grid_vectors = [
+        [grid_spacing, 0.0, 0.0],
+        [0.0, grid_spacing, 0.0],
+        [0.0, 0.0, grid_spacing]
+    ]
+    
+    return grid_points, (nx, ny, nz), origin, grid_vectors
+
+
+def interpolate_esp_to_grid(surface_points, grid_points, method='nearest', max_distance=2.0):
+    """
+    Interpolate surface ESP values to regular grid using nearest neighbor.
+    Points far from surface are set to zero.
+    
+    Args:
+        surface_points: Nx4 array (x, y, z, esp)
+        grid_points: Mx3 array of grid coordinates
+        method: Interpolation method ('nearest', 'linear', or 'cubic')
+        max_distance: Maximum distance for interpolation (points further get 0)
+    """
+    surface_coords = surface_points[:, :3]
+    esp_values = surface_points[:, 3]
+    
+    logger.info(f"Interpolating {len(grid_points)} grid points from {len(surface_points)} surface points...")
+    
+    # Build KD-tree for efficient nearest neighbor search
+    tree = cKDTree(surface_coords)
+    
+    # Find distances to nearest surface point
+    distances, indices = tree.query(grid_points, k=1)
+    
+    # Interpolate using scipy
+    if method == 'nearest':
+        grid_esp = esp_values[indices]
+    else:
+        grid_esp = griddata(surface_coords, esp_values, grid_points, method=method, fill_value=0.0)
+    
+    # Zero out points far from surface
+    grid_esp[distances > max_distance] = 0.0
+    
+    logger.info(f"ESP range on grid: {np.min(grid_esp):.4f} to {np.max(grid_esp):.4f}")
+    logger.info(f"Points within {max_distance:.1f} of surface: {np.sum(distances <= max_distance)}")
+    
+    return grid_esp
+
+
+def write_cube_file(filename, atoms, grid_dims, origin, grid_vectors, values):
+    """Write Gaussian cube file format."""
+    nx, ny, nz = grid_dims
+    
+    with open(filename, 'w') as f:
+        f.write("Cube file: ESP interpolated from xtb surface data\n")
+        f.write("Electrostatic potential (a.u.)\n")
+        
+        # Number of atoms and origin
+        f.write(f"{len(atoms):5d} {origin[0]:12.6f} {origin[1]:12.6f} {origin[2]:12.6f}\n")
+        
+        # Grid dimensions and vectors
+        f.write(f"{nx:5d} {grid_vectors[0][0]:12.6f} {grid_vectors[0][1]:12.6f} {grid_vectors[0][2]:12.6f}\n")
+        f.write(f"{ny:5d} {grid_vectors[1][0]:12.6f} {grid_vectors[1][1]:12.6f} {grid_vectors[1][2]:12.6f}\n")
+        f.write(f"{nz:5d} {grid_vectors[2][0]:12.6f} {grid_vectors[2][1]:12.6f} {grid_vectors[2][2]:12.6f}\n")
+        
+        # Atom coordinates
+        for atom in atoms:
+            atomic_num, x, y, z = atom
+            f.write(f"{atomic_num:5d} {float(atomic_num):12.6f} {x:12.6f} {y:12.6f} {z:12.6f}\n")
+        
+        # ESP values (6 per line)
+        for i in range(0, len(values), 6):
+            line_values = values[i:i+6]
+            f.write(" ".join([f"{v:13.5E}" for v in line_values]) + "\n")
+
+
 class GFN2xTB:
     def __init__(self, molecule: Chem.Mol, ncores: int | None = None):
         assert isinstance(molecule, Chem.Mol), "molecule is not rdkit.Chem.Mol type"
@@ -42,8 +204,11 @@ class GFN2xTB:
             ncores = os.cpu_count()
 
         # Parallelisation
+        # https://xtb-docs.readthedocs.io/en/latest/setup.html#parallelisation
         os.environ['OMP_STACKSIZE'] = '4G'
         os.environ['OMP_NUM_THREADS'] = f'{ncores},1'
+        # OMP_NUM_THREADS=4,2 would mean the outermost parallel region uses 4 threads, 
+        # and any parallel regions nested within it would use up to 2 threads.
         os.environ['OMP_MAX_ACTIVE_LEVELS'] = '1'
         os.environ['MKL_NUM_THREADS'] = f'{ncores}'
         
@@ -457,7 +622,8 @@ $ cp -r xtb-dist/share      /usr/local/ """)
                 elif water == 'alpb':
                     options += ['--alpb', 'water']
                 elif water == 'cpcmx':
-                    logger.warning('optimize with --cpcmx option is not implemented in xtb yet')
+                    logger.warning('CPCM-X not implemented for geometry optimization. '
+                                   'Please use another solvation model for optimization instead.')
 
             if verbose:
                 logger.info(f"optimize() {' '.join(cmd+options)}")
@@ -482,45 +648,51 @@ $ cp -r xtb-dist/share      /usr/local/ """)
 
         # something went wrong if it reaches here
         return SimpleNamespace()
-    
 
-    def esp(self, water: str | None = None, verbose: bool = False) -> None:
-        """Calculate electrostatic potential
+
+    
+    def esp_surface_points(self, water: str | None = None, verbose: bool = False) -> str:
+        """Calculate electrostatic potential surface points data.
         
         Example:
-            v = py3Dmol.view()
-            v.addVolumetricData(dt,
-                                "cube.gz", {
-                                    'isoval': 0.005,
-                                    'smoothness': 2,
-                                    'opacity':.9,
-                                    'voldata': esp,
-                                    'volformat': 'cube.gz',
-                                    'volscheme': {
-                                        'gradient':'rwb',
-                                        'min':-.1,
-                                        'max':.1,
-                                        }
-                                    });
-            v.addModel(dt,'cube')
-            v.setStyle({'stick':{}})
-            v.zoomTo()
-            v.show()
+            def esp_to_rgb(value):
+                import matplotlib.cm as cm
+                import matplotlib.colors as mcolors
+                cmap = cm.get_cmap("seismic")
+                rgb = mcolors.to_hex(cmap(value))
+                return rgb
+
+            # --- Create 3Dmol viewer
+            view = py3Dmol.view(width=600, height=500)
+            view.addModel(xyz_data, "xyz")
+            view.setStyle({'stick': {'radius': 0.2}, 'sphere': {'scale': 0.3}})
+            # --- Add ESP points as colored spheres
+            for xi, yi, zi, vi in zip(x, y, z, norm):
+                color = esp_to_rgb(vi)
+                view.addSphere({
+                    'center': {'x': float(xi), 'y': float(yi), 'z': float(zi)},
+                    'radius': 0.2,
+                    'color': color,
+                    'opacity': 0.9
+                })
+            view.zoomTo()
+            view.show()
         """
+        esp_surface_points_data = ""
         with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
             workdir = Path(temp_dir)
             if verbose:
                 logger.info(f'xtb.optimize workdir= {temp_dir}')
             
             geometry_input_path = workdir / 'geometry.xyz'
-            xtb_esp_dat = workdir / 'xtb_esp_dat'
+            xtb_esp_dat = workdir / 'xtb_esp.dat'
 
             with open(geometry_input_path, 'w') as geometry:
                 geometry.write(self.to_xyz())
 
             cmd = ['xtb', geometry_input_path.as_posix()]
 
-            options = ['--esp', '--gfn', '2']
+            options = ['--sp', '--esp', '--gfn', '2']
 
             if water is not None and isinstance(water, str):
                 if water == 'gbsa':
@@ -533,6 +705,85 @@ $ cp -r xtb-dist/share      /usr/local/ """)
 
             if proc.returncode == 0 and xtb_esp_dat.is_file():
                 with open(xtb_esp_dat, 'r') as f:
-                    pass
+                    esp_surface_points_data = f.read()
+
+        return esp_surface_points_data
+
+
+    def esp_volumetric(self, 
+            water: str | None = None, 
+            grid_spacing: float = 0.6,
+            padding: float = 4.0,
+            max_distance: float = 2.0,
+            method: str = 'nearest',
+            verbose: bool = False) -> tuple[str, float, float]:
+        """Calculate electrostatic potential volumetric data.
         
-        return None
+        Example:
+            import py3Dmol
+
+            view = py3Dmol.view(width=600, height=400)
+            view.addModel(xyz_data, 'xyz')
+            view.setStyle({'stick': {}})
+            view.addVolumetricData(esp_voldata, 'cube', {'isoval': esp_scale_max*0.5, 'color': 'blue', 'opacity': 0.75})
+            view.addVolumetricData(esp_voldata, 'cube', {'isoval': esp_scale_min*0.5, 'color': 'red',  'opacity': 0.75})
+            view.zoomTo()
+            view.show()
+        """
+        esp_voldata = ""
+        esp_scale_min = 0.0
+        esp_scale_max = 0.0
+        with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
+            workdir = Path(temp_dir)
+            if verbose:
+                logger.info(f'xtb.optimize workdir= {temp_dir}')
+            
+            geometry_input_path = workdir / 'geometry.xyz'
+            xtb_esp_dat = workdir / 'xtb_esp.dat'
+            esp_cube = workdir / 'esp.cube'
+
+            with open(geometry_input_path, 'w') as geometry:
+                geometry.write(self.to_xyz())
+
+            cmd = ['xtb', geometry_input_path.as_posix()]
+
+            options = ['--sp', '--esp', '--gfn', '2']
+
+            if water is not None and isinstance(water, str):
+                if water == 'gbsa':
+                    options += ['--gbsa', 'H2O']
+                elif water == 'alpb':
+                    options += ['--alpb', 'water']
+
+            proc = subprocess.run(cmd + options, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
+            # output files: xtb_esp.cosmo, xtb_esp.dat, xtb_esp_profile.dat
+
+            if proc.returncode == 0 and xtb_esp_dat.is_file():                
+                # [1/5] Reading molecular structure
+                atoms = read_xyz_file(geometry_input_path)                
+                
+                # [2/5] Reading surface ESP data
+                surface_points = read_surface_esp(xtb_esp_dat)
+                if verbose:
+                    logger.info(f"Loaded {len(surface_points)} surface points")
+                    logger.info(f"ESP range: {np.min(surface_points[:, 3]):.4f} to {np.max(surface_points[:, 3]):.4f}")
+                
+                # [3/5] Creating regular grid
+                grid_points, grid_dims, origin, grid_vectors = create_regular_grid(surface_points, atoms, grid_spacing, padding)
+                
+                # [4/5] Interpolating ESP to grid
+                grid_esp = interpolate_esp_to_grid(surface_points, grid_points, method, max_distance)
+                
+                # [5/5] Writing cube file
+                write_cube_file(esp_cube, atoms, grid_dims, origin, grid_vectors, grid_esp)
+                esp_scale_min = np.min(grid_esp)
+                esp_scale_max = np.max(grid_esp)
+                
+                if verbose:
+                    logger.info(f"Grid: {grid_dims[0]} x {grid_dims[1]} x {grid_dims[2]} points")
+                    logger.info(f"ESP range: {esp_scale_min:.4f} to {esp_scale_max:.4f} a.u.")
+                    
+                with open(esp_cube, 'r') as f:
+                    esp_voldata = f.read()
+    
+        return (esp_voldata, esp_scale_min, esp_scale_max)
\ No newline at end of file
