@@ -70,6 +70,14 @@ def test_optimize():
     print()
 
 
+def test_esp():
+    mol = testmol.copy()
+    print("GFN2xTB.esp()")
+    outdict = GFN2xTB(mol.confs[0].rdmol).esp_volumetric()
+    print(outdict)
+    print()
+
+
 if __name__ == '__main__':
     test_xtb_wrapper()
     # test_singlepoint()
